<?php
namespace Zeedhi\Framework\Controller;

use Zeedhi\Framework\Routing\Parsers\JSONFormatException;
use Zeedhi\Framework\Annotations\Parser as OpenAPIParser;
use Zeedhi\Framework\Cache\Type\MemcachedImpl;
use Zeedhi\Framework\Process\ProcessManagerImpl;
use Zeedhi\Framework\Process\ProcessImpl;
use Zeedhi\Framework\HTTP\Kernel;
use Zeedhi\Framework\DTO;

class Metadata extends Simple {
    /** @var string constant OPENAPI The openapi process suffix. */
    private const OPENAPI = 'openapi';
    /** @var MemcachedImpl $cache The cache interface to deal with processing. */
    private $cache;

    /**
     * Generates a new openapi.json.
     *
     * @param DTO\Request $request The request.
     * @param DTO\Response $response The response that will show the status.
     * @return void
     */
    public function buildOpenAPI(DTO\Request $request, DTO\Response $response){
        try {
            if($this->notProcessing()){
                $this->doProcess();
            }
        } catch(\Exception $e) {
            $response->setError(new DTO\Response\Error($e->getMessage(), $e->getCode(), $e->getTraceAsString()));
        }
    }

    /**
     * Retrieves the openapi.
     *
     * @param DTO\Request $request The metadata.
     * @param DTO\Response $response The response that will receive openapi.
     * @return void
     */
    public function getOpenAPI(DTO\Request $request, DTO\Response $response){
        try {
            $resource = null;
            if($request->hasParameter('resource')){
                $resource = $request->getParameter('resource');
            }
            $result = OpenAPIParser::getOldJSON(Kernel::$metadataSource, $resource);
            if($result){
                $response->setMetadata($result);
            } else {
                $this->buildOpenAPI($request, $response);
            }
        } catch(\Exception $e) {
            $response->setError(new DTO\Response\Error($e->getMessage(), $e->getCode(), $e->getTraceAsString()));
        }
    }

    /**
     * Retrieves the metadata.
     *
     * @param DTO\Request $request The metadata request.
     * @param DTO\Response $response The response that will receive the metadata loaded.
     */
    public function getMetaData(DTO\Request $request, DTO\Response $response) {
        try {
            $type = $request->hasParameter('type') ? $request->getParameter('type') : 'json';
            $metadata = $this->loadFile($request->getParameter('fileName'), $type);
            $response->setMetadata($metadata);
        } catch(\Exception $e) {
            $response->setError(new DTO\Response\Error($e->getMessage(), $e->getCode(), $e->getTraceAsString()));
        }
    }

    /**
     * Prepares the JSON to be sent.
     *
     * @param mixed $metadata The metadata JSON to be prepared.
     * @return string The prepared json.
     */
    protected function prepareJSON($metadata, $filePath){
        $jsonData = json_decode($metadata, true);
        $path = Kernel::$metadataSource.DIRECTORY_SEPARATOR.$filePath;
        if ($jsonData === null) {
            throw JSONFormatException::invalidJSON($path, json_last_error_msg());
        }

        return $jsonData;
    }

    /**
     * Loads the metadata file.
     *
     * @param  string $filePath The path to the file.
     * @param  string $type The type of the file.
     * @return string The loaded file.
     */
    protected function loadFile($filePath, $type){
        $data = file_get_contents(Kernel::$metadataSource.DIRECTORY_SEPARATOR.$filePath.'.'.$type);

        if(strtolower($type) === 'json') {
            $data = $this->prepareJSON($data, $filePath);
        }

        return $data;
    }

    /*
     * Searches for present openapi.json generation processes.
     * @codeCoverageIgnore
     *
     * @return boolean
     */
    private function notProcessing() : bool {
        $memcached = new \Memcached;
        $memcached->addServer(ProcessManagerImpl::HOST, ProcessManagerImpl::PORT);
        $this->cache = new MemcachedImpl($memcached);
        /** @method $filterCallback Callback to filter cached keys. */
        $filterCallback = function($id){
            return strpos($id, self::OPENAPI) !== FALSE;
        };
        return empty($this->cache->filterKeys($filterCallback));
    }

    /**
     * Creates OpenAPI process.
     * @codeCoverageIgnore
     */
    private function createOpenAPIProcess(){
        $processCallback = function(){
            file_put_contents(
                Kernel::$metadataSource.DIRECTORY_SEPARATOR.'openapi.json',
                OpenAPIParser::getJSON()
            );
        };
        $process = new ProcessImpl($processCallback, self::OPENAPI);
        return $process;
    }


    /**
     * Run openAPI generation process.
     * @codeCoverageIgnore
     */
    private function generateOpenAPI(){
        $process = $this->createOpenAPIProcess();
        $processId = $process->getId();
        $pm = new ProcessManagerImpl($this->cache);
        $pm->addProcess($process);
        $pm->run($processId);
        $pm->discard($processId);
    }

    /**
     * Starts process of openapi.json generation.
     * @codeCoverageIgnore
     */
    private function doProcess(){
        Kernel::forceRequestDisconnect(202);
        $this->generateOpenAPI();
    }
}