<?php
namespace Zeedhi\Framework\DependencyInjection;

use Symfony\Component\DependencyInjection\ContainerBuilder;
use Symfony\Component\Config\FileLocator;
use Symfony\Component\DependencyInjection\Definition;
use Symfony\Component\DependencyInjection\Exception\BadMethodCallException;
use Symfony\Component\DependencyInjection\Loader\PhpFileLoader;
use Symfony\Component\DependencyInjection\Loader\XmlFileLoader;
use Symfony\Component\DependencyInjection\Loader\YamlFileLoader;

/**
 * Class InstanceManager
 *
 * Dependency injector that can generate instances using class definitions and configured instance parameters
 *
 * @package Zeedhi\Framework\DependencyInjection
 */
class InstanceManager {
    public const YAML = 'YAML';
    public const XML = 'XML';
    public const PHP = 'PHP';

    /**
     * @var InstanceManager
     */
    protected static $instanceManager;
    /**
     * @var \Symfony\Component\DependencyInjection\ContainerBuilder
     */
    protected $container;

    /**
     * @var \Symfony\Component\DependencyInjection\Loader\FileLoader
     */
    protected $loader;

    /**
     * Constructor
     */
    private function __construct() {
        $this->container = new ContainerBuilder();
    }

    /**
     * Get an container the dependency
     *
     * @return \Symfony\Component\DependencyInjection\ContainerBuilder
     */
    public function getContainer() : ContainerBuilder {
        return $this->container;
    }

    /**
     * Gets a service.
     *
     * @param string $id The service identifier
     *
     * @return object The associated service
     *
     * @throws \Exception See \Symfony\Component\DependencyInjection\ContainerBuilder::get documentation
     */
    public function getService(string $id) {
        return $this->container->get($id);
    }

    /**
     * Sets a service.
     *
     * @param string $id      The service identifier
     * @param object $service The service instance
     *
     * @return InstanceManager
     */
    public function registerService(string $id, $service) : InstanceManager {
        $this->container->set($id, $service);
        return $this;
    }

    /**
     * Gets a parameter.
     *
     * @param string $name The parameter name
     *
     * @return mixed  The parameter value
     *
     * @throws \InvalidArgumentException if the parameter is not defined
     *
     */
    public function getParameter(string $name) {
        return $this->container->getParameter($name);
    }


    /**
     * Sets a service definition.
     *
     * @param string     $id         The service identifier
     * @param Definition $definition A Definition instance
     *
     * @return Definition the service definition
     *
     * @throws BadMethodCallException When this ContainerBuilder is frozen
     *
     * @api
     */
    public function setDefinition(string $id, Definition $definition) : Definition {
        return $this->container->setDefinition($id, $definition);
    }


    /**
     * Sets a parameter.
     *
     * @param string $name  The parameter name
     * @param mixed  $value The parameter value
     *
     * @return InstanceManager
     */
    public function setParameter(string $name, $value) : InstanceManager {
        $this->container->setParameter($name, $value);
        return $this;
    }

    /**
     * Compiles the container.
     *
     * This method loads the annotation classes and passes the
     * container to compiler passes whose job is to manipulate
     * and optimize the container.
     */
    public function compile() {
        $this->loadAnotationFromFile();
        $this->container->compile();
    }

    /**
     * Loads a dependency file.
     *
     * @param string $file
     * @param string $type
     */
    public function loadFromFile(string $file, $type = InstanceManager::XML) {
        $dirName = dirname($file);
        $fileName = basename($file);
        switch ($type) {
            case self::YAML:
                $this->loader = new YamlFileLoader($this->container, new FileLocator($dirName));
                $this->loader->load($fileName);
                break;
            case self::XML:
                $this->loader = new XmlFileLoader($this->container, new FileLocator($dirName));
                $this->loader->load($fileName);
                break;
            case self::PHP:
                $this->loader = new PhpFileLoader($this->container, new FileLocator($dirName));
                $this->loader->load($fileName);
        }
    }

    /**
     *  Create an instance of InstanceManager
     *
     * @return static
     */
    public static function getInstance() : InstanceManager {
        if (!self::$instanceManager) {
            self::$instanceManager = new static();
        }
        return self::$instanceManager;
    }

    /**
     * Checks if container is compiled.
     *
     * @return boolean
     */
    public function isCompiled() : bool {
        return $this->container->isCompiled();
    }

    /**
     * Returns all definitions.
     *
     * @return Definition[]
     */
    public function getDefinitions() : array {
        return $this->container->getDefinitions();
    }

    /**
     * Loads a dependency annotation from file.
     */
    private function loadAnotationFromFile() {
        $ds = DIRECTORY_SEPARATOR;
        $filePath = __DIR__.$ds.'..'.$ds.'..'.$ds.'..'.$ds.'..'.$ds.'annotations.xml';
        $xml = simplexml_load_string(file_get_contents($filePath));
        foreach($xml->services->service as $annotation) {
            $id = strval($annotation->attributes()->id);
            $class = strval($annotation->attributes()->class);
            $object = new $class();
            $this->registerService($id, $object);
            $this->forceLoadById($id);
        }
    }

    /**
     * Forces the container to create an instance of the class.
     *
     * @param string serviceId The service identifier.
     */
    public function forceLoadById(string $serviceId) {
        $this->getService($serviceId);
    }
}